##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
PDGClassifier <- setRefClass(
  Class = "PDGClassifier",
  fields = list(
    variablesDataset = "list",
    database = "Database",
    classIndex = "numeric",
    collapseRate = "numeric",
    mergeRate = "numeric",
    method = "character",
    maxIntervals="numeric",
    densityType="character",
    nTermsDensity = "numeric",
    maxTermsDensity = "numeric"
  )
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
PDGClassifier$accessors(c("classIndex", "variablesDataset","maxIntervals","densityType",
							"database"))

##################################################################################################
############################################# METHODS ############################################
##################################################################################################

# Initializes the PDG classifier with a database and the position for the class variable.
#
# Args:
#   database: A object of type Database.
#   classIndex: the column in the database in which the class variable is located
#               (non mandatory argument). If the argument is not given the class variable will
#               be located in the last position.
#   collapseRate: A number k in the formula nCases = k * R(Xi), where R(Xi) is the number of
#                 states in the variable Xi and nCases is the minimum number of cases to be
#                 allowed to learn the parameters without collapsing nodes. By default this
#                 number is set to 0.5.
#   mergeRate: A number between 0 and 1 indicating the percentage of parameters to be checked
#              for merging (non mandatory argument). By default this argument is set to 0.5.
#
#   method: the method to build the classifier:
#           "ranked": The variables are ranked by mutual information and they are connected as:
#                     C -> X1 -> X2 -> X3 -> ...
#           "rankedCR": The variables are ranked by mutual information and then each Xi is
#                             connected to the leaf that maximizes the CR of the classifier wrt to a
#                              test database.
#           "CR": In each step a non included variable Xj is connected to a leaf variable Xi
#                      satisfying that <Xi, Xj> = arg max CR(Xi,Xj,pdg,train,test). This is
#                      computationally the hardest option. See Nielsen et. al (2009) - Algorithm 2.
#
#   maxIntervals: The maximum number of intervals in which the continuous parents are
#                                  split.
#
#   densityType: the type of density ("MOP" or "MTE").
#
#   maxTermsDensity: the maximum number of terms to be fit in the density.
#
# Returns:
#   The initialized PDGClassifier object.

PDGClassifier$methods (
  initialize = function(database, classIndex=NULL, collapseRate=NULL,
                        mergeRate=NULL, method=NULL, maxIntervals =NULL,
                        densityType=NULL, maxTermsDensity=NULL){

   database <<- database

   if(is.null(classIndex)) classIndex <<- database$getNumberOfVars()
   else classIndex <<- classIndex

   if(is.null(collapseRate)) collapseRate <<- 5
   else collapseRate <<- collapseRate


   if(is.null(mergeRate)) mergeRate <<- 0.5
   else
     if(mergeRate<0 || mergeRate>1) stop("\nMerge rate must be between 0 and 1. ")
      else
        mergeRate <<- mergeRate

   if(is.null(method)) method <<- "ranked"
   else method <<- method

   if(is.null(maxIntervals)) maxIntervals <<- 3
   else maxIntervals <<- maxIntervals

   if(is.null(densityType)) densityType <<- "MOP"
   else densityType <<- densityType

   if(!is.null(maxTermsDensity))
   		maxTermsDensity <<- maxTermsDensity


   for (i in 1:database$getNumberOfVars()) {
     if(database$getDataType(i) == "continuous")
       variablesDataset[[length(variablesDataset)+1]] <<- ContinuousVariableNode$new(database$getNameVar(i),database$getMin(i),database$getMax(i))
     else
       variablesDataset[[length(variablesDataset)+1]] <<- CategoricalVariableNode$new(database$getNameVar(i),database$getStates(i))
   }
  }
)

##################################################################################################

# Gets the root of the PDG classifier. This variable is located in the position given by the field
# classIndex in the list of variables.
#
# Args:
#
# Returns:
#   The root of PDGClassifier object.

PDGClassifier$methods (
    getRoot = function () {
      variablesDataset[[classIndex]]
    }
)

##################################################################################################

# Gets the variable located in a given position in the list of variables.
#
# Args:
#   i: Position of the varible.
#
# Returns:
#   The variable in position i in the list of variables.

PDGClassifier$methods (
  getVariableAt = function (i) {
    variablesDataset[[i]]
  }
)

##################################################################################################

# Gets the name of the variables of the PDG
#
# Args:
#
# Returns:
#   A vector of variable names
#

PDGClassifier$methods (
  getVariableNames = function() {
    vNames=c()
    for (var in getVariablesDataset()){
      vNames = c(vNames,var$getName())
    }
    return(vNames)
  }
)

##################################################################################################

# Gets the variable with the name given as argument.
#
# Args:
#   name: Name of the variable.
#
# Returns:
#   The corresponding variable.

PDGClassifier$methods (
  getVariableByName = function (name) {
    index = which(getVariableNames()==name)
    return(getVariableAt(index))
  }
)

##################################################################################################

# Removes a parameter leaf from the model considering the impact on its predecessors. Note that
# the parameter to remove does not have successors.
#
# Args:
#   param: Parameter node to be removed.
#
# Returns:
#   The updated PDG after the removal.

PDGClassifier$methods (
  removeParameterLeafFromPDG = function (param) {
    # Remove param as successor from all its predecessors
    lapply(param$getPredecessors(), function(pred){ pred$removeSuccessor(param) })
    # Remove the predecessors of param
    param$removePredecessors()
  }
)

##################################################################################################

# Removes a variable from the model.
#
# Args:
#   var: Variable to be removed.
#
# Returns:
#   The updated PDG after the removal.

PDGClassifier$methods (
  removeVariableNodeFromPDG = function (var) {
    var$removeParameters()
    var$removeChildren()
    parent = var$getParent()[[1]]
    parent$removeChild(var)
  }
)

##################################################################################################

# Recursive method to get the variables included in a set of trees with roots stored
# in a given queue
#
# Args:
#   queue: Queue with a set of root variables.
#
# Returns:
#   The variables included in the model ordered by the method Breadth First Search.

PDGClassifier$methods (
  searchVariablesBFS = function(queue){
    if(!queue$empty()) {
      node = queue$get()
      a = lapply(node$getChildren(), function(x){queue$insert(x)})
      return (c(node,searchVariablesBFS(queue)))
    }
  }
)

##################################################################################################

# Gets the variables included in the PDG classifier using the method Breadth First Search. Note
# that these variables are not neccessarely those included in the database. For that purpose use
# the method getVariablesDataset().
#
# Args:
#
# Returns:
#   The variables included in the model ordered by the method Breadth First Search.

PDGClassifier$methods (
  getVariables = function(){
    searchVariablesBFS(Queue$new(getRoot()))
  }
)

##################################################################################################

# Gets the variables of the dataset
#
# Args:
#
# Returns:
#   The variables included in the dataset.

PDGClassifier$methods (
  getVariablesDataset = function(){
    variablesDataset
  }
)


##################################################################################################

# Gets the parameters of the model. First obtain the variables and then extract the parameters
# included in each variable.
#
# Args:
#
# Returns:
#   The parameters of the PDG classifier.

PDGClassifier$methods (
  getParameters = function(){

    parameters = list()
    for (v in getVariables()){
      for (i in 1:v$numParameterNodes()){
         parameters[[length(parameters)+1]] = v$getParameterAt(i)
      }
    }
    return(parameters)
  }
)


##################################################################################################

# Shows the structure of the PDG: name of the variables and links between them.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

PDGClassifier$methods (
   showStructure = function() {
    a= lapply(getVariables(), function(x) {
      a = lapply(x$getChildren(),
             function(y){
               cat(paste(x$getName()," -> ",y$getName(), sep=""))
               cat("\n")
               a = lapply(x$getParameters(),
                      function(z) {
                        cat(paste("    ", z$getName(),"-> ( "), sep="")
                        a= lapply(z$getSuccessors(),
                               function(w) {
                                 if(w$getVar()$getName()==y$getName())
                                   cat(paste(w$getName(), " ", sep=""))
                        })
                        cat(")\n")
                })
      })
    })
   }
)

##################################################################################################

# Shows the parameters of the PDG with their name, links between them, distribution and
# reached cases.
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

PDGClassifier$methods (
    showParameters = function() {
      cat("_______________________________________________________\n")
      a=lapply(getVariables(), function(x) {
        a=lapply(x$getParameters(), function(y) {
          y$show();
          cat("_______________________________________________________\n")

        })
      })
  }
)

##################################################################################################

# Shows relevant information about the PDG classifier node: variables, structure and parameters.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

PDGClassifier$methods (
  show = function() {
    cat("-------------------------------------------------------\n")
    cat("--------------------- PDG STRUCTURE--------------------\n")
    cat("-------------------------------------------------------\n")
    showStructure()
    cat("-------------------------------------------------------\n")
    cat("------------------- PDG DISTRIBUTIONS -----------------\n")
    cat("-------------------------------------------------------\n")
    showParameters()
  }
)

##################################################################################################

# This recursive method counts the number of parameter nodes bellow a given variable.
#
# Args:
#   var: A variable of the model
#
# Returns:
#   The number of parameters bellow a given variable.

PDGClassifier$methods (
  countParametersBellowVar = function(var)  {
    n=var$numParameterNodes()
    if(var$isLeaf())
      return (n)
    else {
      for (i in 1:var$numChildren()){
        n = n + countParametersBellowVar(var$getChildAt(i))
      }
    }
    return (n)
  }
)

##################################################################################################

# Counts the number of parameter nodes of the PDG.
#
# Args:
#
# Returns:
#   The number of parameters of the PDG.

PDGClassifier$methods (
  numParameterNodes = function()  {
    countParametersBellowVar(getRoot())
  }
)

##################################################################################################

# Counts the number of descendants of a variable, that is, the number of variables bellow it
# in the PDG.
#
# Args:
#
# Returns:
#   The number of descendants for var.

PDGClassifier$methods (
  numDescendants = function(var)  {
    n=1
    if(var$isLeaf())
      return (1)
    else {
      for (i in 1:var$numChildren()){
        n = n + numDescendants(var$getChildAt(i))
      }
    }
    return (n)
  }
)

##################################################################################################

# Counts the number of descendants of a variable, that is, the number of variables bellow it
# in the PDG.
#
# Args:
#
# Returns:
#   The number of descendants for var.

PDGClassifier$methods (
  numVariables = function()  {
    length(getVariables())
  }
)

##################################################################################################

# Counts the number leaf variables in the model.
# in the PDG.
#
# Args:
#
# Returns:
#   The number of leaf variables in the model.

PDGClassifier$methods (
  numLeafVariables = function()  {
    num=0
    for (var in getVariables()){
      if(var$isLeaf())
        num = num + 1
    }
    return(num)
  }
)

##################################################################################################

# Plots the distribution of the parameters into files.
#
# Args:
#
# Returns:
#   One file for each parameter is created.

PDGClassifier$methods (
  plotParameters = function(pathFolder)  {
      a = lapply(getParameters(),function(x){x$plot(paste(pathFolder, "/", x$getName(),".pdf",sep=""))})
 }
)

##################################################################################################

# Print the undo information.
#
# Args:
#   undoInfo: The undo information list
# Returns:
#   The undo information
#

PDGClassifier$methods (
  printUndoInfo = function(undoInfo,v,w)  {
    cat("-------------------------------------------------------\n")
    cat("------------------ UNDO INFORMATION --------------------\n")
    cat("-------------------------------------------------------\n")


    cat("\n PARAMETER:", v$getName())
    vPredList = undoInfo$v$vPredecessors
    vPosList =  undoInfo$v$vPositionsSucc
    vFreePos = undoInfo$v$vFreePositions
    vNumSucc = undoInfo$v$vNumSuccessors
    n = length(vPredList)
    for(i in 1:n){ cat("\n   ", vPredList[[i]]$getName(),
                       "   Positions:", vPosList[[i]],
                       " Free positions:",vFreePos[[i]],
                      " Number os successors:", vNumSucc[[i]])}


    cat("\n\n PARAMETER:", w$getName())
    wPredList = undoInfo$w$wPredecessors
    wPosList =  undoInfo$w$wPositionsSucc
    wFreePos = undoInfo$w$wFreePositions
    wNumSucc = undoInfo$w$wNumSuccessors
    n = length(wPredList)
    for(i in 1:n){ cat("\n   ", wPredList[[i]]$getName(),
                       "   Positions:", wPosList[[i]],
                       " Free positions:",wFreePos[[i]],
                      " Number os successors:", wNumSucc[[i]])}

   cat("-------------------------------------------------------\n")
   cat("-------------------------------------------------------\n")
   cat("-------------------------------------------------------\n")


    if(length(vPredList)!=length(vPosList) || length(wPredList)!=length(wPosList))
          stop("\n Error")
  }
)
##################################################################################################

# Compute the classification rate of the model given a set of predicted values and test values.
#
# Args:
#
# Returns:
#   The classification rate between 0 and 1.

PDGClassifier$methods (
 getClassificationRate = function(testValues, predictedValues)  {

    if(length(testValues)!=length(predictedValues))
      stop("Test (", length(testValues), ") and predicted values (", length(predictedValues), ") length difer")

    cmp=(testValues==predictedValues)
    return(length(cmp[cmp==TRUE])/length(testValues))
  }
)

##################################################################################################

# Returns the parameter nodes in the PDG reached by a given assignment.
#
# Args:
#   assignment: An assigment.
#
# Returns:
#   The parameter nodes reached by the assignment.

PDGClassifier$methods (
  reachedNodes = function(assignment)  {
    rootParameterNode = .self$getRoot()$getParameterAt(1)
    rootParameterNode$reachedNodes(assignment)
  }
)

##################################################################################################

# Remove the parameter nodes that are not reached from a parameter node.
#
# Args:
#   param: the parameter from which the method begins.
# Returns:
#   The PDG is updated after removing the unreached parameter nodes.

PDGClassifier$methods (
  removeUnreachedParametersFrom = function(param)  {
    var = param$getVar()
    var$removeParameter(param)
    #cat("\n            Removing", param$getName(), "from the variable", var$getName())

    for(succ in param$getSuccessors()){
      #cat("\n            Removing", param$getName(), "from the list of predecessor of", succ$getName())
      succ$removePredecessor(param)
      succReached=FALSE
      for(pred in succ$getPredecessors()){
        if(!identical(pred,param)){
          succReached=TRUE
        }
      }
      if(succReached ==FALSE){
          removeUnreachedParametersFrom(succ)
      }
    }

  }
)

##################################################################################################

# Returns the unique parameter node belonging to a variable from a list of reached nodes.
#
# Args:
#   var: A variable.
#   reachedNodes: A set of reached nodes.
#
# Returns:
#   A parameter node satisfying the condition.

PDGClassifier$methods (
  reach = function(var, reachedNodes)  {
    for(nodo in reachedNodes){
      #if(var$equals(nodo$getVar())){
      if(identical(nodo$getVar(),var)){
        return(nodo)
      }
    }
    return(NULL)
  }
)

##################################################################################################

# Update the information stored in the variables concerning their parameters and
# predecessors, as they may be inconsistent after removing unreached nodes recursively during
# the merge operation.
#
# Args:
# v: the parameter node from which the update will start.
#
# Returns:
#   The modified PDG.

PDGClassifier$methods (
  updateParametersInVariables = function(v) {

    # The current parameter is added to the variable if it was not added yet.
    if(v$getVar()$hasParameter(v)==FALSE){
      #cat("\n        Adding parameter", v$getName(),"to variable",v$getVar()$getName())
      v$getVar()$addParameter(v)
      for(succ in v$getSuccessors()){
        #cat("\n        Adding parameter", v$getName(),"as predecessor of parameter",succ$getName())
        succ$addPredecessor(v)
      }
    }
    for(succ in v$getSuccessors()){
      updateParametersInVariables(succ)
    }
  }
)


##################################################################################################

# Collects the information stored in v and w needed to undo a unsuccessful merge operation.
#
# Args:
# v: the first parameter node involved in the merge operation.
# w: the second parameter node involved in the merge operation.
#
# Returns:
#   A list with 2 components:
#     - v: a list of predecessors of v and a list of vector of positions in which v is stored
#          as successor of each predecessor.
#     - w: a list of predecessors of w and a list of vector of positions in which v is stored
#          as successor of each predecessor.

PDGClassifier$methods (
  collectUndoInfo = function(v,w) {

    #------------------------------------------------------------------------
    vPositionsSucc = list()
    vPredecessors= list()

    for(pred in unique(v$getPredecessors())){
      vPredecessors[[length(vPredecessors)+1]] = pred
      vPos = pred$getPositionsSucc(v)
      vPositionsSucc[[length(vPositionsSucc)+1]] = vPos
    }
    #------------------------------------------------------------------------
    wPositionsSucc = list()
    wPredecessors= list()
    for(pred in unique(w$getPredecessors())){
      wPredecessors[[length(wPredecessors)+1]] = pred
      wPos = pred$getPositionsSucc(w)
      wPositionsSucc[[length(wPositionsSucc)+1]] = wPos
    }
    #------------------------------------------------------------------------
    wFreePositions= list()
    vFreePositions = list()
    vNumSuccessors=list()
    wNumSuccessors=list()
    j=1
    for (pred_v in unique(v$getPredecessors())){
      vFreePositions[[j]] = vPositionsSucc[[j]]
      vNumSuccessors[[j]] = pred_v$numSuccessors()
      j=j+1
    }
    j=1
    for (pred_w in unique(w$getPredecessors())){
      wFreePositions[[j]] = wPositionsSucc[[j]]
      wNumSuccessors[[j]] = pred_w$numSuccessors()

      j=j+1
    }

    j = 1
    for (pred_v in unique(v$getPredecessors())){
      k=1
      for(pred_w in unique(w$getPredecessors())){
        if(identical(pred_v,pred_w)){
          vFree = setdiff(1:pred_v$numSuccessors(), union(vPositionsSucc[[j]],wPositionsSucc[[k]]))
          vFreePositions[[j]] = setdiff(1:pred_v$numSuccessors(),vFree)
          wFreePositions[[k]] = setdiff(1:pred_w$numSuccessors(),vFree)
        }
        k=k+1
      }
      j=j+1
    }

    #------------------------------------------------------------------------
    vList=list(vPredecessors=vPredecessors,vPositionsSucc=vPositionsSucc, vFreePositions = vFreePositions, vNumSuccessors=vNumSuccessors)
    wList=list(wPredecessors=wPredecessors,wPositionsSucc=wPositionsSucc, wFreePositions = wFreePositions, wNumSuccessors=wNumSuccessors)
    #------------------------------------------------------------------------
    undoInfo=list(v=vList, w=wList)
    return(undoInfo)
  }
)

##################################################################################################

# Merges two parameters of this variable. The order of the argument is very important since
# merge(v,w) is different from merge(w,v)!!!!
#
# Args:
#   v: A parameter node.
#   w: A parameter node.
#
# Returns:
#   The modified PDG.
#

PDGClassifier$methods (
  mergeParameters = function(v,w)  {

    if(!identical(v$getVar(),w$getVar()))  stop("Parameters must belong to the same variable.")

    parent = v$getVar()$getParent()[[1]]
    var = v$getVar()


    if(var$isCategorical()){
      vNew = CategoricalParameterNode$new(name=v$getName())
    }
    else {
      vNew = ContinuousParameterNode$new(name=v$getName(),maxIntervals=getMaxIntervals())
      # The cutpoints of the new parameters are equals to the one in v
      vNew$setCutPoints(v$getCutPoints())
    }

    #cat("\n      ___________________________________________________")
    #cat("\n      Merging parameters", v$getName(), "and", w$getName(), "into", vNew$getName(), "(new) ...")
    #cat("\n      ___________________________________________________")
    #cat("\n      Creating the new parameter", vNew$getName(), "(new)")
    #cat("\n      The new parameter", vNew$getName(), "belongs now to variable", var$getName(),"...")
    vNew$getDistribution()$setVar(var)

    #cat("\n      Merging cases from", v$getName(),"and",w$getName(), "into", vNew$getName(),"(new)...")
    vNew$getReachedData()$mergeCases(v$getReachedData())
    vNew$getReachedData()$mergeCases(w$getReachedData())

    #cat("\n       Re-learn the potential")
    #potNew = ProbabilityVector$new(var=var)
    #potNew$learning(vNew$getReachedData()$getColumnData(var))
    #cat("\n      Assigning the averaged potential of", v$getName(), "and", w$getName(), "to",vNew$getName(), "(new) ...")
    potNew = var$getAveragedPotential(v,w)
    vNew$setDistribution(potNew)

    #cat("\n      Successors of",v$getName(), "are now assigned to", vNew$getName(), "(new)")
    vNew$setSuccessors(v$getSuccessors())

    #cat("\n      For each successor of", vNew$getName(), "(new):")
    lapply(vNew$getSuccessors(), function(succ) {
      #cat("\n         Removing", v$getName(), "(old) as predecessor of", succ$getName(), "...")
      succ$removePredecessor(v)
      #cat("\n         Adding", vNew$getName(), "(new) as predecessor of", succ$getName(), "...")
      succ$addPredecessor(vNew)
    })

    for(pred in v$getPredecessors()){
      #cat("\n      For predecessor",pred$getName())
      pos = pred$getPositionsSucc(v)[1]
      #cat("\n         Successor position of", v$getName(), "(old) in the list of successor of", pred$getName(),":", pos)
      #cat("\n         Inserting", vNew$getName(),"(new) in position", pos, "as successor of", pred$getName())
      pred$addSuccessorAt(vNew,pos)
      #cat("\n         Adding", pred$getName(), "as predecessor of", vNew$getName())
      vNew$addPredecessor(pred)
      #cat("\n         Removing", v$getName(), "(old) as successor of", pred$getName())
      pred$removeSuccessor(v)
    }

    for(pred in w$getPredecessors()){
      #cat("\n      For predecessor",pred$getName())
      pos = pred$getPositionsSucc(w)[1]
      #cat("\n         Successor position of", w$getName(), "in the list of successor of", pred$getName(),":", pos)
      #cat("\n         Inserting", vNew$getName(),"(new) in position", pos, "as successor of", pred$getName())
      pred$addSuccessorAt(vNew,pos)
      #cat("\n         Adding", pred$getName(), "as predecessor of", vNew$getName())
      vNew$addPredecessor(pred)
      #cat("\n         Removing", w$getName(), "(old) as successor of", pred$getName())
      pred$removeSuccessor(w)
    }

    #cat("\n      Adding new parameter", vNew$getName(), "to the variable", var$getName())
    var$addParameter(vNew)

    #cat("\n      Removing old parameter", v$getName(), "from the variable", var$getName())
    var$removeParameter(v)
    # Parameter w is removed inside the method below.

    #cat("\n      Removing unreached parameters from node", w$getName(),"...")
    # We only check the parameters unreached from w. Those outgoing from v are always reached
    removeUnreachedParametersFrom(w)
    #cat("\n")
  }
)

##################################################################################################

# Undo the changes after the merge operation. The parameters given as arguments have the
# information needed to go back to the original PDG structure.
#
# Args:
#   v: the first parameter node used in the merge operation.
#   w: the second parameter node used in the merge operation.
#   undoInfo: a list with two attributes:
#        vPositions: a vector with the positions of v as successor of its predecessors.
#        wPositions: a vector with the positions of w as successor of its predecessors.
#
# Returns:
#   The original PDG before the merge operation.
#

PDGClassifier$methods (
  undoMerge = function(v,w,undoInfo) {

    #cat("\n      ___________________________________________________")
    #cat("\n      Undo merge operation over parameters",v$getName(), "and", w$getName())
    #cat("\n      ___________________________________________________")

    # Remove the new parameter added to the model located in the last position in the list
    # of parameters of the variable.
    # --------------------------------------------------------------------------------------
    var = v$getVar()
    lastPos = var$numParameterNodes()
    paramToRemove = var$getParameterAt(lastPos)

    #cat("\n      Removing", paramToRemove$getName(), "from the variable", var$getName())
    var$removeParameterAt(lastPos)
    # --------------------------------------------------------------------------------------
    # Remove the parameter from the list of successors for every predecessor.
    lapply(paramToRemove$getPredecessors(), function(pred) {
      #cat("\n      Removing", paramToRemove$getName(), "as successor of", pred$getName())
      pred$removeSuccessor(paramToRemove)
    })
    # --------------------------------------------------------------------------------------
    # Remove the parameter from the list of predecessor for every successor
    lapply(paramToRemove$getSuccessors(), function(succ) {
      #cat("\n      Removing", paramToRemove$getName(), "as predecessor of", succ$getName())
      succ$removePredecessor(paramToRemove)
    })
    # --------------------------------------------------------------------------------------
    # Clean the successors and predecessors of the new parameter
    paramToRemove$removeSuccessors()
    paramToRemove$removeSuccessors()
    # --------------------------------------------------------------------------------------
    # Add the former parameters to the variable again.
    #cat("\n      Adding", v$getName(), "to the variable", var$getName())
    var$addParameter(v)
    #cat("\n      Adding", w$getName(), "to the variable", var$getName())
    var$addParameter(w)
    # --------------------------------------------------------------------------------------
    #printUndoInfo(undoInfo,v,w)

    for(i in 1:length(unique(v$getPredecessors()))){
      pred = undoInfo$v$vPredecessors[[i]]
      if(pred$numSuccessors()!=undoInfo$v$vNumSuccessors[[i]]){
        for(pos in undoInfo$v$vFreePositions[[i]]){
          #cat("\n   Adding an empty successor to predecessor", pred$getName(), "at position", pos)
          pred$addSuccessorAt(CategoricalParameterNode$new(name="-"),pos)
        }
      }
    }
    for(i in 1:length(unique(w$getPredecessors()))){
      pred = undoInfo$w$wPredecessors[[i]]
      if(pred$numSuccessors()!=undoInfo$w$wNumSuccessors[[i]]){
        for(pos in undoInfo$w$wFreePositions[[i]]){
          #cat("\n   Adding an empty successor to predecessor", pred$getName(), "at position", pos)
          pred$addSuccessorAt(CategoricalParameterNode$new(name="-"),pos)
        }
      }
    }

    # --------------------------------------------------------------------------------------
    #cat("\n      Updating the former the information about successors of predecessors of",v$getName(),"and",w$getName())
    #cat("\n------------------------ ",v$getName()," ---------------------------\n")
    vPredList = undoInfo$v$vPredecessors
    vPosList = undoInfo$v$vPositionsSucc
    i=1
    for(pred in vPredList){
      #cat("\n    Predecessor =",pred$getName(),"   ")
      #cat(sort(vPosList[[i]]))
      #cat("\n       Successors before:",pred$getSuccessorNames())

      for(pos in vPosList[[i]]){
       # cat("\n       Set successor",v$getName(),"at position", pos)
       # pred$addSuccessorAt(v,pos)
        pred$setSuccessorAt(v,pos)
      }
      #cat("\n       Successors after:",pred$getSuccessorNames())

      i=i+1
    }
    #cat("\n------------------------ ",w$getName()," ---------------------------\n")
    wPredList = undoInfo$w$wPredecessors
    wPosList = undoInfo$w$wPositionsSucc
    i=1
    for(pred in wPredList){
      #cat("\n    Predecessor =",pred$getName(), "   ")
      #cat("\n       Successors before:",pred$getSuccessorNames())

      for(pos in wPosList[[i]]){
        #cat("\n       Set successor",w$getName(),"at position", pos)
        pred$setSuccessorAt(w,pos)
      }
      #cat("\n       Successors after:",pred$getSuccessorNames())
      i=i+1
    }
    # --------------------------------------------------------------------------------------
    #cat("\n      Updating the former the information about predecessors of successors of",v$getName(),"and",w$getName())
    #---------------------------------------------------------------------------------------
    lapply(v$getSuccessors(), function(succ) {
      #cat("\n        Adding", v$getName(),"as predecessor of",succ$getName())
      succ$addPredecessor(v)
    })
    lapply(w$getSuccessors(), function(succ) {
      #cat("\n        Adding", w$getName(),"as predecessor of",succ$getName())
      succ$addPredecessor(w)
    })
    #-----------------------------------------------------------------------------------------------
    # Unreached nodes were recursively removed from the variables during the merge, so we
    # need to update the information about parameters included in the variables as well as their
    # predecessors. This information is taken from the successors of the parameters.
    v0 = getRoot()$getParameterAt(1)
    updateParametersInVariables(v0)
    #cat("\n")
  }
)

###################################################################################################

# Carries out the merge operation bottom up from the leaves. On each PDG level a number of merges
# (depending on a mergeRate parameter) are checked .
#
# Args:
#   test: a dataset for testing the CR after each merge operation is carried out.
# Returns:
#   The modified PDG.

PDGClassifier$methods (
  mergeParametersBottomUp = function(test)  {

    #cat("\n=====================================================")
    #cat("\n       MERGING PARAMETERS BOTTOM UP")
    #cat("\n                Rate:", mergeRate)
    #cat("\n=====================================================")

    #cat("\n Computing CR ... ")
    .self$predict(test)
    CR = .self$getClassificationRate(test$getColumnData(getVariableAt(classIndex)), .self$predict(test))
    #cat("\nCR = ", CR)
    # Computes the reversed order of variables previously obtained by BFS (except the root)
    # in order to explore the tree from the leaves to the root. The root is not included in the
    # list of variables to check.
    variablesBottomUp = rev(getVariables()[-1])

    for (var in variablesBottomUp) {
      #cat("\n____________________________________________________")

      #cat("\nVariable:", var$getName())
      # Remember that var$getParents() is a list due to implementation aspects, but it must have
      # only 1 variable

      parent = var$getParent()[[1]]

      # The operation is only performed in variables with categorical parents.
      #if(parent$isCategorical()){

        numParameters = var$numParameterNodes()
        #cat("\n   Number of parameters:", numParameters)
        numChecks = (numParameters*mergeRate)%/%2
        #cat("\n   Number of merges to be checked:", numChecks)

        if (numChecks>0){
          #####################
          set.seed(1) # BORRAR
          ######################
          indexes = sample(1:numParameters,numChecks*2)
          #cat("\n   Random parameter positions: ",indexes)

          parametersToMerge = var$getParameters()[indexes]
          #print(parametersToMerge)
          for (i in seq(1,length(indexes),2)){
            #cat("\n      Indexes:",indexes[i] , "and", indexes[i+1])
            v = parametersToMerge[[i]]
            w = parametersToMerge[[i+1]]
            #cat("\n      Parameters:", v$getName(), "and", w$getName())


            undoInfo = collectUndoInfo(v,w)
            #cat("\nANTES DE MERGE\n")
            #showStructure()
            mergeParameters(v,w)
            #cat("\nDESPUES DE MERGE\n")
            #showStructure()
            #cat("\n      End merge parameters:", v$getName(), "and", w$getName(), "\n")

            # La siguiente linea se puede borrar
            #.self$checkValidity()

            # --------------------------------------------------------------------
            #cat("\n      Checking classification accuracy for the new merge ...")
            predictedValues = .self$predict(test)
            testValues = test$getColumnData(getVariableAt(classIndex))
            newCR = .self$getClassificationRate(testValues, predictedValues)

            #cat("\n      Old CR:", CR)
            #cat("\n      New CR:", newCR)
            if (newCR>=CR){
              #cat("\n      Successful merge.")
              CR = newCR
            }
            else{
              #cat("\n      Merge does not improve the CR. Undo the changes")
              undoMerge(v,w,undoInfo)


             # cat("\nDESPUES DE UNDO\n")
              #showStructure()


              # Las siguientes 3 lineas se pueden borrar. Simplemente valen para comprobar
              # que la operación undo devuelve un pdg con igual CR que el orginial antes de
              # aplicar merge.
             # afterMergeCR = .self$getClassificationRate(test$getColumnData(getVariableAt(classIndex)), .self$predict(test))
              #cat("\nCA=",CR, "- AfterUndoCR:",afterMergeCR)
            #  if(CR!=afterMergeCR) stop("Merge operation is not working ...")

              # La siguiente linea se puede borrar
             # .self$checkValidity()
            }
          }
        }
     # }
    }
    #cat("\n=====================================================")
    #cat("\n        END MERGING PARAMETERS BOTTOM UP")
    #cat("\n=====================================================\n")

    # La siguiente linea se puede borrar
    #.self$checkValidity()
  }
)

###################################################################################################

# Collapses the parameters included in this leaf variable that satisfy the criteria
# reachedData < collapseRate·|R(X)|. The procedure ends when the previous criteria is satisfied or the number
# of parameter nodes in this variable is lower than 2.
#
# Args:
#   var: A leaf variable.
#
# Returns:
#   The modified PDG.

PDGClassifier$methods (
  collapse = function(var)  {

    if(var$isCategorical()) {
      threshold = .self$collapseRate*var$numStates()
    }

    #cat("\n__________________________________________________\n")
    #cat(" Collapsing parameters in variable", var$getName())
    #cat("\n__________________________________________________")

   # .self$checkValidity()

    repeat {
      if(var$isContinuous()) {
          vNumParams=c()
          for(param in var$getParameters()){
            vNumParams = c(vNumParams, param$numParameters())
          }
          threshold = .self$collapseRate*vNumParams
      }

      v = var$numReachedCasesParameters()
      #cat("\n Reached cases:",v)
      #cat("\n     Threshold:", threshold)
      numParametersToCollapse = length(v[v<threshold])
      #cat("\n Remaining parameters to be collapsed:", numParametersToCollapse)
      if(numParametersToCollapse < 1) {
        break
      }
      else { # at least one parameter to be collapsed
        i = which(v==min(v))[1]
        v[i] = Inf
        j = which(v==min(v))[1]
        #cat("\n       Positions to collapse:", i,"and",j)
        v1 = var$getParameterAt(i)
        v2 = var$getParameterAt(j)

        var$collapseParameters(v1,v2, maxTerms=maxTermsDensity)

      }
    }
    #.self$checkValidity()
  }
)

##################################################################################################

# Compute the joint probability distribution for a given assignment by multiplying the
# probabilities obtained after evaluating the assignment in the PDG.
#
# Args:
#   assignment: An assignment.
#
# Returns:
#   A value representing the joint probability distribution of the assignment in the PDG.

PDGClassifier$methods (
  jointDistribution = function(assignment)  {

    #cat("\n      Computing reached nodes ...")
    reachedNodes = .self$reachedNodes(assignment)
    #cat("\n      Reached nodes computed. \n")
    vProbs=c()
#     cat("\n--------------\n")
#     print(assignment)
#     cat("\n--------------\n")
#     print(pdg)
#     cat("\n--------------\n")
#     print(reachedNodes)
#     cat("\n--------------\n")

    for(i in 1:assignment$size()) {
      var = assignment$getVarAt(i)
      node = .self$reach(var,reachedNodes)
      if (!is.null(node)) {
        value  = assignment$getValue(var)
        #cat("\n   Variable:", var$getName())
        prob = node$evaluate(value)
        #cat("\n   Parameter:",node$getName(), "- P(", value,") = ", prob)
        if(is.nan(prob)) { stop("Probability is NaN") }
        vProbs = c(vProbs,prob)
      }
    }
    return(prod(vProbs))
  }
)

##################################################################################################

# Classify an evidence. All the features must be instantiated. For each state c of the class,
# the joint distribution for the assignment {assignment,C=c} is computed and the value c with
# maximum joint distribution value is returned. Note that class variable C can be located anywhere
# in the assignment.
#
# Args:
#   assignment: A configuration of feature values.
#
# Returns:
#   The value of the class that maximizes P(C|X1=x1, ..., Xn=xn)

PDGClassifier$methods (
  classify = function(featuresAssignment)  {
    root = .self$getRoot()
    joints=c()

    featuresAssignment$addElement(root,"")
   # cat("\nCompute the argmax(P(C|x)")
    for(state in root$getStates()){
      #cat("\n   C :", state)
      featuresAssignment$setValueAt(state,featuresAssignment$size())
      joint = jointDistribution(featuresAssignment)
      #cat("\n   Joint distribution:", joint)
      joints=c(joints,joint)
    }
    featuresAssignment$removeElementAt(featuresAssignment$size())
    posMax = which.max(joints)
    return(root$getState(posMax))
  }
)

#################################################################################################
# Return a vector of predicted values for the class variable, one for each record of
# a given dataset. This task is carried out using the package doParallel, meaning that
# the available cores in the machine will be used in parallel to compute the vector of predictions.
#
# Args:
#   test: A Database
#
# Returns:
#   A vector of predicted values.

PDGClassifier$methods (
  predict = function(test)  {

#       predictedValues <- foreach(i=1:test$getNumberOfCases(), .combine=c) %dopar% {
#                           featuresAssignment = MixedAssignment$new()
#                           features = getVariablesDataset()[-classIndex]
#                           featuresAssignment$loadFromDataFrameRow(features,test$getData()[i,-c(classIndex)])
#                           .self$classify(featuresAssignment)
#       }

      #--------------------------------------------------------------------------------
     # SIN PARALELIZAR
    predictedValues=c()
    for(i in 1:test$getNumberOfCases()){
      featuresAssignment = MixedAssignment$new()
      #cat("\n---------------------------------------------------------")
      #cat("\nLoading features assignment from data row", i,"\n\n")
      features = getVariablesDataset()[-classIndex]
      #print(features)
      featuresAssignment$loadFromDataFrameRow(features, test$getData()[i,-c(classIndex)])
      #print(featuresAssignment)
      #cat("\n Classifying instance...")
      prediction = .self$classify(featuresAssignment)
      #cat("\nPrediction=", prediction)
      predictedValues= c(predictedValues,prediction)
    }
      #--------------------------------------------------------------------------------
      return(predictedValues)
  }
)


##################################################################################################

# Gets the probability value after classifying an evidence. All the features must be instantiated.
# For each state c of the class,the joint distribution for the assignment {assignment,C=c} is
# computed and the maximum joint distribution value for c is returned. Note that class variable C
# can be located anywhere in the assignment.
#
# Args:
#   assignment: A configuration of feature values.
#
# Returns:
#   The probability value maximizing P(C|X1=x1, ..., Xn=xn)

PDGClassifier$methods (
  degree = function(featuresAssignment)  {
    root = .self$getRoot()
    joints=c()

    featuresAssignment$addElement(root,"")
    # cat("\nCompute the argmax(P(C|x)")
    for(state in root$getStates()){
      #cat("\n   C :", state)
      featuresAssignment$setValueAt(state,featuresAssignment$size())
      joint = jointDistribution(featuresAssignment)
      #cat("\n   Joint distribution:", joint)
      joints=c(joints,joint)
    }
    featuresAssignment$removeElementAt(featuresAssignment$size())
    posMax = which.max(joints)
    return(joints[posMax]/sum(joints))
  }
)

#################################################################################################
# Return a vectorwith the probability values associated to the predicted values for the class
# variable, one for each record of a given dataset. This task is carried out using the package
# doParallel, meaning that the available cores in the machine will be used in parallel to compute
# the vector of predictions.
#
# Args:
#   test: A Database
#
# Returns:
#   A vector of probability associated to the predicted values for the class.

PDGClassifier$methods (
  score = function(test)  {

    #       scoredValues <- foreach(i=1:test$getNumberOfCases(), .combine=c) %dopar% {
    #                           featuresAssignment = MixedAssignment$new()
    #                           features = getVariablesDataset()[-classIndex]
    #                           featuresAssignment$loadFromDataFrameRow(features,test$getData()[i,-c(classIndex)])
    #                           .self$degree(featuresAssignment)
    #       }

    #--------------------------------------------------------------------------------
    # SIN PARALELIZAR
    scoredValues=c()
    for(i in 1:test$getNumberOfCases()){
      featuresAssignment = MixedAssignment$new()
      #cat("\n---------------------------------------------------------")
      #cat("\nLoading features assignment from data row", i,"\n\n")
      features = getVariablesDataset()[-classIndex]
      #print(features)
      featuresAssignment$loadFromDataFrameRow(features, test$getData()[i,-c(classIndex)])
      #print(featuresAssignment)
      #cat("\n Classifying instance...")
      mapprediction = .self$degree(featuresAssignment)
      #cat("\nMaximum Probability Prediction=", mapprediction)
      scoredValues= c(scoredValues,mapprediction)
    }
    #--------------------------------------------------------------------------------
    return(scoredValues)
  }
)

##################################################################################################

# Gets the class probability distribution after classifying an evidence. All the features must be
# instantiated. For each state c of the class,the joint distribution for the assignment {assignment,C=c} is
# computed andreturned. Note that class variable C can be located anywhere in the assignment.
#
# Args:
#   assignment: A configuration of feature values.
#   states: a boolean variable, if TRUE class states are include in the output
#
# Returns:
#   The probability distribution P(C|X1=x1, ..., Xn=xn)

PDGClassifier$methods (
  classDistribution = function(featuresAssignment, states = TRUE)  {
    root = .self$getRoot()
    joints=c()

    featuresAssignment$addElement(root,"")
    # cat("\nCompute the argmax(P(C|x)")
    for(state in root$getStates()){
      #cat("\n   C :", state)
      featuresAssignment$setValueAt(state,featuresAssignment$size())
      joint = jointDistribution(featuresAssignment)
      #cat("\n   Joint distribution:", joint)
      joints=c(joints,joint)
    }
    featuresAssignment$removeElementAt(featuresAssignment$size())
    #posMax = which.max(joints)
    if (states == TRUE) {
       names(joints) <- root$getStates()
    }
    joints <- joints/sum(joints)
    return (joints)
  }
)

#################################################################################################
# Return a matrix with the class probability distribution associated to each record of a given
# dataset. This task is carried out using the package doParallel, meaning that the available
# cores in the machine will be used in parallel to compute
# the vector of predictions.
#
# Args:
#   test: A Database
#
# Returns:
#   A matrix with the class probability distribution associated to each instance in the Database.

PDGClassifier$methods (
  scores = function(test)  {

    #       scoredValues <- foreach(i=1:test$getNumberOfCases(), .combine=c) %dopar% {
    #                           featuresAssignment = MixedAssignment$new()
    #                           features = getVariablesDataset()[-classIndex]
    #                           featuresAssignment$loadFromDataFrameRow(features,test$getData()[i,-c(classIndex)])
    #                           .self$degree(featuresAssignment)
    #       }

    #--------------------------------------------------------------------------------
    # SIN PARALELIZAR
    scoredValues=c()
    for(i in 1:test$getNumberOfCases()){
      featuresAssignment = MixedAssignment$new()
      #cat("\n---------------------------------------------------------")
      #cat("\nLoading features assignment from data row", i,"\n\n")
      features = getVariablesDataset()[-classIndex]
      #print(features)
      featuresAssignment$loadFromDataFrameRow(features, test$getData()[i,-c(classIndex)])
      #print(featuresAssignment)
      #cat("\n Classifying instance...")
      classPrediction = .self$classDistribution(featuresAssignment, FALSE)
      #cat("\nMaximum Probability Prediction=", mapprediction)
      scoredValues= c(scoredValues, classPrediction)
    }
    #--------------------------------------------------------------------------------
    root = .self$getRoot()
    states = root$getStates()
    classDistributions <-  matrix(scoredValues, ncol=length(states), byrow = TRUE)
    colnames (classDistributions) <- states
    return(classDistributions)
  }
)

##################################################################################################

# Adds a new variable to the model by connecting it to another previously included.
#
# Args:
#   existingVariable: A variable already included in the model.
#   newVariable: The new variable to be inserted.
#
# Returns:
#  The PDG with the new variable.

PDGClassifier$methods (
  addNewVariableFromExistingVariable = function(existingVariable, newVariable) {
    Xi = newVariable
    Xj = existingVariable
    Xj$addChild(Xi)
    #cat(paste("\n(", Xj$getName()," -> ", Xi$getName(),")",sep=""))

    #-------------------------------------------------------------------------------------
    if(Xj$isCategorical() && Xi$isCategorical()){
      #cat(" Categorical -> Categorical\n")
      for(paramXj in Xj$getParameters()){ # For each parameter node of Xj
        for(state in Xj$getStates()) { # For each state of variable Xj
          #cat("\n                   ", paste(Xi$getName(),".", state,sep=""))
          v = CategoricalParameterNode$new()
          paramXj$addSuccessor(v)
          v$addPredecessor(paramXj)
          v$setReachedData(paramXj$getReachedData()$selectDataFromVariableState(Xj,state))
          pot = ProbabilityVector$new(var=Xi)

          dataToLearn = v$getReachedData()$getColumnData(Xi)
          #cat("\n",Xj$getName(),"= ",state,"\n       Cases:", length(dataToLearn))
          pot$learning(dataToLearn)
          v$setDistribution(pot)
          Xi$addParameter(v)
        }
        #cat("\n")
      }
    }
    #-------------------------------------------------------------------------------------
    if(Xj$isCategorical() && Xi$isContinuous()){
      #cat(" Categorical -> Continuous\n")
      for(paramXj in Xj$getParameters()){ # For each parameter node of Xj

        for(state in Xj$getStates()) { # For each state of variable Xj
          v = ContinuousParameterNode$new(maxIntervals=getMaxIntervals())
          paramXj$addSuccessor(v)
          v$addPredecessor(paramXj)
          v$setReachedData(paramXj$getReachedData()$selectDataFromVariableState(Xj,state))
          dataToLearn = v$getReachedData()$getColumnData(Xi)
          #cat("\n",Xj$getName(),"= ",state,"\n       Cases:", length(dataToLearn))
          pot = MoTBFDensity$new(var=Xi, type=getDensityType())
          v$setDistribution(pot$fitDensity(dataToLearn,maxTerms = maxTermsDensity))
          Xi$addParameter(v)
        }
        #cat("\n")
      }
    }
    #-------------------------------------------------------------------------------------
    if(Xj$isContinuous() && Xi$isCategorical()){
      #cat(" Continuous -> Categorical\n")
      for(paramXj in Xj$getParameters()){ # For each parameter node of Xj
        paramXj$addVectorOfCutPointsEF()

        for(i in 1:(getMaxIntervals())) { # For each interval

          #cat("\n                   ", paste(Xi$getName(),".", state,sep=""))
          v = CategoricalParameterNode$new()
          paramXj$addSuccessor(v)
          v$addPredecessor(paramXj)
          a = paramXj$getCutPoints()[[1]][i]
          b = paramXj$getCutPoints()[[1]][i+1]
          v$setReachedData(paramXj$getReachedData()$selectDataFromVariableInterval(Xj, a,b))
          dataToLearn = v$getReachedData()$getColumnData(Xi)
          #cat("\n",Xj$getName(),"= [",a,b,"]\n       Cases:", length(dataToLearn))
          pot = ProbabilityVector$new(var=Xi)
          pot$learning(dataToLearn)
          v$setDistribution(pot)
          Xi$addParameter(v)
        }
        #cat("\n")
      }
    }
    #-------------------------------------------------------------------------------------
    if(Xj$isContinuous() && Xi$isContinuous()){
      #cat("  Continuous -> Continuous\n")
      for(paramXj in Xj$getParameters()){ # For each parameter node of Xj

        #cat("CUTPOINTS BEFORE:"); print(paramXj$getCutPoints())
        paramXj$addVectorOfCutPointsEF()
        #cat("CUTPOINTS AFTER:"); print(paramXj$getCutPoints())


        for(i in 1:(getMaxIntervals())) { # For each interval
          v = ContinuousParameterNode$new(maxIntervals=getMaxIntervals())
          paramXj$addSuccessor(v)
          v$addPredecessor(paramXj)
          a = paramXj$getCutPoints()[[1]][i]
          b = paramXj$getCutPoints()[[1]][i+1]
          dataParamXj = paramXj$getReachedData()
          v$setReachedData(dataParamXj$selectDataFromVariableInterval(Xj,a,b))
          dataToLearn = v$getReachedData()$getColumnData(Xi)
          #cat("\n",Xj$getName(),"= [",a,b,"]\n       Cases:", length(dataToLearn))

          pot = MoTBFDensity$new(var=Xi, type=getDensityType())
          #cat("\nDENSITY TYPE:", pot$getType())
          v$setDistribution(pot$fitDensity(dataToLearn,maxTerms = maxTermsDensity))
          Xi$addParameter(v)
        }
        #cat("\n")
      }
    }


    #-------------------------------------------------------------------------------------
    if (collapseRate>0.0) {
      #cat("\n Number of parameters of ", newVariable$getName()," before collapsing:",newVariable$numParameterNodes())

       collapse(newVariable)#puedo poner Xi

       #cat("\n Number of parameters of ", newVariable$getName()," after collapsing:",newVariable$numParameterNodes())

    }

  }
)

##################################################################################################

# Sets the names to the parameter nodes of the model. This method is called after building
# the whole model. The names are: (v1, v2,...,vn)
#
# Args:
#
# Returns:
#   The modified with the names in the parameter nodes.

PDGClassifier$methods (
 setParameterNames = function() {
    i = 0
    for (param in getParameters()){
      param$setName(paste("v",i,sep=""))
      i = i+1
    }
 }
)



##################################################################################################

# Gets the features.
#
# Args:
#
# Returns:
#
# A vector of features.

PDGClassifier$methods (
  getFeatures = function () {
    getVariablesDataset()[-classIndex]
  }
)



##################################################################################################

# Gets the number of features.
#
# Args:
#
# Returns:
#
# The number of features.

PDGClassifier$methods (
  numFeatures = function () {
    length(getFeatures())
  }
)

##################################################################################################

# Gets the variables sorted by their mutual information with the class C.
#
# Args:
#
# Returns:
#
# A vector of variables, ordered decreasingly by mutual information.

PDGClassifier$methods (
  getFeaturesOrderedByMI = function () {
    MIs = c()
    vPositions = c()

    db = getDatabase()
    C = db$getColumnDataAt(classIndex)

    for(i in 1:db$getNumberOfVars()){
      if(i!=classIndex){
        Xi = db$getColumnDataAt(i)
        if(getVariableAt(i)$isContinuous()){
          # "infotheo::" to avoid conflict with method discretize from package 'bnlearn'
          Xi = infotheo::discretize(Xi)
        }
        MIs = c(MIs, infotheo::mutinformation(Xi,C))
        vPositions = c(vPositions,i)
      }
    }
    df=data.frame(MIs=MIs,vPositions=vPositions)
    df = df[with(df, order(-MIs)), ]
    orderedVariables = getVariablesDataset()[df$vPositions]

    #cat("\nOrdered features using I(Xi,C): ")
    #for(i in 1:length(orderedVariables)) {
    #  cat("\n   ", df[i,1], "(",orderedVariables[[i]]$getName(),")")
    #}

    return(orderedVariables)
  }
)

##################################################################################################

# Remove the vector of cutpoints located in the last position for every parameter the variable
# given as parameter.
#
# Args:
#  Xj: a Variable
#
# Returns:
#
# The current PDG is uptdated.
#

PDGClassifier$methods (
  removeLastVectorOfCutPoints = function (Xj) {

  if(Xj$isContinuous()){
     for(p in Xj$getParameters()){
       p$setCutPoints(p$getCutPoints()[-p$numCutPoints()])
     }
  }
 }
)

##################################################################################################

# Learn a PDG classifier by including the features according to a ranking given by I(Xi,C).
# The variables are connected in sequence, C -> X1 -> X2 -> X3 -> ...
#
# Args:
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learnRanked = function() {

    db = getDatabase()

    cat("_____________________________________________________\n")
    db$show()
    cat("_____________________________________________________")

    # First, the root variable and parameter v0 are created
    C = getRoot()
    #cat("\nLearning the categorical root parameter ...")
    pot = ProbabilityVector$new(var=C)
    pot$learning(db$getColumnData(C))
    v0 = CategoricalParameterNode$new()
    v0$setReachedData(db)
    v0$setDistribution(pot)
    C$setParameters(c(v0))
    cat("\n   ",pot$getParameters())
    cat("\n_____________________________________________________")
    orderedFeatures = getFeaturesOrderedByMI()
    lastIncludedVar = C

    for(i in 1:length(orderedFeatures)) {
      #cat("\n_____________________________________________________")
      newVar = orderedFeatures[[i]]
      .self$addNewVariableFromExistingVariable(lastIncludedVar,newVar)

      lastIncludedVar = newVar
      setParameterNames()
    }

  }
)

##################################################################################################

# Learn a PDG classifier using a rankedCR method as follows: the variables are ranked by their
# mutual information with the class and then each Xi is connected to that leaf that maximizes
# the CR of the classifier wrt to a test database.
#
# Args:
#   train: the train dataset used to estimate the potentials.
#   test: the test dataset used to compute the CR.
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learnRankedCR = function(train,test) {
    #cat("\n=====================================================")
    #cat("\n=====================================================")
    #cat("\n======== RANKED-CR APPROACH FOR LEARNING ============")
    #cat("\n=====================================================")
    #cat("\n=====================================================")

    # First, the root variable and parameter v0 are created
    C = getRoot()
    pot = ProbabilityVector$new(var=C)
    pot$learning(train$getColumnData(C))
    v0 = CategoricalParameterNode$new()
    v0$setReachedData(train)
    v0$setDistribution(pot)
    C$setParameters(c(v0))

    orderedFeatures = getFeaturesOrderedByMI()
    includedVariables = list(C)

    for (Xi in orderedFeatures) {
      CRs = c()
      for(Xj in includedVariables) {# For each included variable Xj
        #cat("\n_____________________________________________________")
         #cat("\nTrying new variable:")
        .self$addNewVariableFromExistingVariable(Xj,Xi)

        #cat("\n_____________________________________________________")
        #showStructure()
        #cat("\n_____________________________________________________")

        #cat("\n Obtaining test values ...")
        testValues = test$getColumnData(C)
        #cat("\n Computing predicted values ...")
        predictedValues = .self$predict(test)
        #cat("\n Get classification rate ...")

        #print(testValues)
        #print(predictedValues)
        CR = .self$getClassificationRate(testValues, predictedValues)
        #cat("\n"); showStructure()
        #cat("\n   Classification accuraccy:", CR)
        CRs = c(CRs, CR)
        # Undo the changes
        #cat("\n Undo the changes ...")
        lapply(Xi$getParameters(), function(param){.self$removeParameterLeafFromPDG(param)})
        .self$removeVariableNodeFromPDG(Xi)
        .self$removeLastVectorOfCutPoints(Xj)

      }
      pos = which.max(CRs)
      Xj = includedVariables[[pos]]

      #cat("\n=====================================================")
      #cat("\n            Including ", Xj$getName(), "->", Xi$getName())
      #cat("\n=====================================================")

      .self$addNewVariableFromExistingVariable(Xj,Xi)
      includedVariables[[length(includedVariables)+1]] <- Xi
    }
  }
)

##################################################################################################

# Learn a PDG classifier using the following method: in each step a non included variable
# Xj is connected to a leaf variable Xi satisfying that <Xi, Xj> = arg max CR(Xi,Xj,pdg,train,test).
# See Nielsen et. al (2009) - Algorithm 2.
#
# Args:
#   train: the train dataset used to estimate the potentials.
#   test: the test dataset used to compute the CR.
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learnCR = function(train,test) {
    #cat("\n=====================================================")
    #cat("\n=====================================================")
    #cat("\n=========== CR-BASED APPROACH FOR LEARNING ==========")
    #cat("\n=====================================================")
    #cat("\n=====================================================")
    # First, the root variable and parameter v0 are created
    C = getRoot()
    pot = ProbabilityVector$new(var=C)
    pot$learning(train$getColumnData(C))
    v0 = CategoricalParameterNode$new()
    v0$setReachedData(train)
    v0$setDistribution(pot)
    C$setParameters(c(v0))

    includedVariables = list(C)
    nonIncludedVariables = getVariablesDataset()[-c(classIndex)]

    repeat
    {
      CRs = c()
      for(Xj in includedVariables) {# For each included variable Xj
        for(Xi in nonIncludedVariables) { # For each non-included variable Xi
          #cat("\n_____________________________________________________")
          #cat("\n Adding new variable ...")
          .self$addNewVariableFromExistingVariable(Xj,Xi)
          #cat("\n Obtaining test values ...")
          testValues = test$getColumnData(C)
          #cat("\n Computing predicted values ...")
          predictedValues = .self$predict(test)
          #cat("\n Get classification rate ...")
          CR = .self$getClassificationRate(testValues, predictedValues)
          #cat("\n"); showStructure()
          #cat("\n   Classification accuraccy:", CR)
          CRs = c(CRs, CR)

          #cat("\n Undo the changes ...")
          lapply(Xi$getParameters(), function(param){.self$removeParameterLeafFromPDG(param)})
          .self$removeVariableNodeFromPDG(Xi)
          .self$removeLastVectorOfCutPoints(Xj)

        }
      }
      # The position of the max CR give us information about the link Xj->Xi to include in the PDG
      pos = which.max(CRs)
      jpos = (pos-1)%/%length(nonIncludedVariables) +1
      ipos = (pos-1)%%length(nonIncludedVariables)+1
      Xj = includedVariables[[jpos]]
      Xi = nonIncludedVariables[[ipos]]

      #cat("\n=====================================================")
      #cat("\n            Including ", Xj$getName(), "->", Xi$getName())
      #cat("\n=====================================================")

      .self$addNewVariableFromExistingVariable(Xj,Xi)

      nonIncludedVariables = nonIncludedVariables[-c(ipos)]
      includedVariables[[length(includedVariables)+1]] <- Xi

      if(length(nonIncludedVariables)==0) break
    }
  }
)


##################################################################################################

# Learn a PDG classifier with a NB structure.
#
# Args:
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learnNB = function() {
    #cat("\n=====================================================")
    #cat("\n=====================================================")
    #cat("\n=============== PDG with a NB structure =============")
    #cat("\n=====================================================")
    #cat("\n=====================================================\n")

    db = getDatabase()

    #cat("_____________________________________________________\n")
    #db$show()
    #cat("_____________________________________________________")

    # First, the root variable and parameter v0 are created
    C = getRoot()
    #cat("\nLearning the categorical root parameter ...")
    pot = ProbabilityVector$new(var=C)
    pot$learning(db$getColumnData(C))
    v0 = CategoricalParameterNode$new()
    v0$setReachedData(db)
    v0$setDistribution(pot)
    C$setParameters(c(v0))

    features = getFeatures()

    for(i in 1:numFeatures()) {
      .self$addNewVariableFromExistingVariable(C,features[[i]])
    }
    setParameterNames()
  }
)


##################################################################################################

# Learn a PDG classifier using a TAN structure given by the Chow Liu algorithm.
#
# Args:
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learnChowLiu = function() {

    db = getDatabase() # new modification

    # We discretize the data for learning the TAN structure
    discDB = Database$new(data=db$getData())
    if (!discDB$isDiscretized())
        discDB$discretizeDatasetEF(3)

    # First, the root variable and parameter v0 are created
    C = getRoot()
    pot = ProbabilityVector$new(var=C)
    pot$learning(db$getColumnData(C))
    v0 = CategoricalParameterNode$new()
    v0$setReachedData(db)
    v0$setDistribution(pot)
    C$setParameters(c(v0))

    # TAN
    nameClass = db$getNameVar(classIndex)
    data = data.frame(lapply(discDB$getData(), factor))
    nameFeatures = colnames(data)[1:(classIndex-1)]


    # TAN with mutual information coefﬁcients for the Chow-Liu algorithm
    tan = bnlearn::tree.bayes(data, nameClass, nameFeatures)


    numLinksTAN = length(tan$arcs[,1])
    tree = tan$arcs[(numFeatures()+1):numLinksTAN,]
    treeRoot = getVariableByName(tree[1,1])
    .self$addNewVariableFromExistingVariable(C,treeRoot)
    numPendingArcs = length(tree[,1])

    while (numPendingArcs>0) {
        for(i in 1:numPendingArcs) {
            parentName = tree[i,1]
            childName = tree[i,2]

            Xj = getVariableByName(parentName)
            Xi = getVariableByName(childName)

            # check if the variable can be inserted to the current PDG
            flag = FALSE
            for (v in getVariables()){
                if (v$getName() == parentName){
                    .self$addNewVariableFromExistingVariable(Xj,Xi)
                    tree = data.frame(tree[-c(i),])
                    numPendingArcs = numPendingArcs-1
                    flag = TRUE
                    break;
                }
            }
            if(flag) break
        }
    }
  }
)










##################################################################################################

# Learns a PDG model from a dataset.
#
# Args:
#   db: a dataset for learning
#
# Returns:
#   The PDG model

PDGClassifier$methods (
  learn = function() {

    #cat("\n=====================================================")
    #cat("\n     COLLAPSING PARAMETERS: ", collapseRate,"*", "R(Xi)")
    #cat("\n=====================================================")
    #cat("\n             MERGE RATE: ", mergeRate)
    #cat("\n=====================================================\n")

    db = getDatabase()
    # The second argument is the seed, remove it if randomness is required.
    datasets=db$divideIntoTrainAndTest(0.2, 1)
    train = datasets$dbtrain
    test = datasets$dbtest

    switch(method,
           "ranked"   = { learnRanked() },
           "rankedCR" = { learnRankedCR(train, test) },
           "CR"       = { learnCR(train,test) },
           "NB"       = { learnNB() },
           "ChowLiu"  = { learnChowLiu() }
    )



    setParameterNames()

    if(mergeRate>0){
      mergeParametersBottomUp(test)
    }
    #cat("\n\nModel learned.\n")
 }
)

##################################################################################################

# Gets the parameter at a given position
#
# Args:
#   i: the position of the parameter
#
# Returns:
#   The parameter at position i.
#
PDGClassifier$methods (
  getParameterAt = function(i) {
      getParameters()[[i]]
  }
)

##################################################################################################

# Gets the size of the model (number of parameters).
# size(pdg) = sum(size(parameter_i)),  where
# size(parameter_i) = number of parameters of the potential -1
# For continuous parameters, we also add the number of cutpoints in case of having children.
#
# Args:
#
# Returns:
#   The size of the model.
#

PDGClassifier$methods (
  size = function() {
    sum(sapply(getParameters(),function(p){p$size()}))
  }
)






##################################################################################################

# Samples a set of values for the variables bellow a parameter given as an argument
#
# Args:
#
# Returns:
#   A vector of values
#

PDGClassifier$methods (
  sampleValuesFromNode = function(v, values){

    pot = v$getDistribution()
    value = pot$simulateValue()

    #cat("\nSampled value for variable:", v$getVar()$getName(), ":", value)

    #Compute the column index of this variable
    i = which(getVariableNames() == v$getVar()$getName())
    values[i]=value

    assignment = MixedAssignment$new(variables = c(v$getVar()), values=list(value))
    reachedNodes = v$reachedNodes(assignment)

    for(nextNode in reachedNodes[-1]){
      values = sampleValuesFromNode(nextNode,values)
    }
    return(values)
  }
)

##################################################################################################

# Samples a data instance from the PDG
#
# Args:
#
# Returns:
#   A simulated data instance
#

PDGClassifier$methods (
  sampleDataInstance = function(){
    v0 = getRoot()$getParameterAt(1)
    return(sampleValuesFromNode(v0, vector(length=numVariables())))
  }
)

##################################################################################################

# Samples a dataset from the PDG.
#
# Args:
#
# Returns:
#   A csv file with the simulated values is saved.
#

PDGClassifier$methods (
  sampleFromPDG = function(size, csvFile) {
    simulatedData = matrix(ncol=numVariables(), nrow=0)
    colnames(simulatedData) = getVariableNames()
    for (i in 1:size){
      simulatedData = rbind(simulatedData,sampleDataInstance())
    }
    write.table(simulatedData, file=csvFile, sep=",", quote=F, row.names=F)
  }
)

##################################################################################################

# Check the validity of a PDG in terms of the following issues:
# 1. All variables except the root must have exactly one parent.
# 2. All children of a variable must have this variable as a parent.
# 3. Root variable cannot have any parent.
# 4. Non-root parameters must have some predecessor.
# 5. Non-leaf parameters must have some successor.
# 6. Probabilities in parameters must sum up to one.
# 7. The number of successors of a parameter which are associated to a same children must be equal
#    to the number of states of the variable to which the parameter belongs.
# 8. Each successor of a parameter must include this parameter as predecessor.
# 9. Each predecessor of a parameter must include this parameter as successor.
# 10. The sum of reached cases in leaves for a same variable must match the reached cases in the root.
#
# Args:
#
# Returns:
#   Report any inconsistency on the PDG structure.
#

PDGClassifier$methods (
  checkValidity = function() {
    sumCasesParameterLeaves = 0
    totalPredecessors = 0
    totalSuccessors = 0



    for(var in getVariables()){
      #--------------------------------------------------------------------------
      if (!identical(getRoot(),var) && (length(var$getParent())!=1)){
        stop("The variable ", var$getName(), " does not have exactly one parent")
      }
      #--------------------------------------------------------------------------
      for(child in var$getChildren()){
        if(!identical(child$getParent()[[1]],var))
          stop("Inconsistency in relation parent-children:", var$getName(),"->",child$getName())
      }
      #--------------------------------------------------------------------------
      if (identical(getRoot(),var) && (length(var$getParent())!=0)){
        stop("Root variable cannot have any parent")
      }
      #cat("\nVAR:",var$getName())

      for(param in var$getParameters()){

        totalPredecessors = totalPredecessors+param$numPredecessors()
        totalSuccessors = totalSuccessors+param$numSuccessors()


        if(!param$isLeaf() && param$isContinuous()){
          if(length(param$getCutPoints())!=param$numChildren()){
            stop("Variable:", var$getName(),". A continuous parameter must have cutpoints.")
          }
        }

        if(!identical(param$getVar(),getRoot()) && param$numPredecessors()==0){
          stop("\n Non-root parameter must have predecessors")
        }
        if(!param$isLeaf() && param$numSuccessors()==0){
          stop("\n Non-leaf parameter must have successors")
        }

        #--------------------------------------------------------------------------
        if(param$isLeaf()){
          #cat("\n              Reached cases in the leaf ", param$getName(), ":",  param$numReachedCases())
          sumCasesParameterLeaves = sumCasesParameterLeaves + param$numReachedCases()
        }

        #--------------------------------------------------------------------------
        for(succ in param$getSuccessors()){
          if (!succ$isPredecessor(param)){
            stop("Inconsistency information stored in the successor ",
                 succ$getName(), " and the predecessor ", param$getName())
          }
        }
        #--------------------------------------------------------------------------
        for(pred in param$getPredecessors()){
          if (!pred$isSuccessor(param)) {
            stop("Inconsistency information stored in the successor",
                 param$getName(), "and the predecessor", pred$getName())
          }
        }
        #--------------------------------------------------------------------------
        if (param$isContinuous() && !param$isLeaf()){
          if(param$numCutPoints()==0)
            stop("Parameter ", param$getName(), " must have cutpoints since it has successors.")
        }
      }

      if(var$isCategorical()){
        #--------------------------------------------------------------------------
        # Hay que redondear para que sea == 1 :(
        if(round(sum(param$getDistribution()$getParameters()),10)!=1){
          stop("Probabilities in parameter ", param$getName(),
               " of variable ", var$getName()," does not sum up to one")
        }
        #--------------------------------------------------------------------------
        for (child in var$getChildren()){
          #cat("\n      CHILD:", child$getName())
          if(param$numSuccessorsVariable(child)!=var$numStates()){
            stop("Number of successors of parameter ", param$getName(),
                 " in variable ", child$getName(), " is not equal to the
                 number of states of variable ", var$getName())
          }
        }
      }




    }
    if(totalPredecessors!=totalSuccessors){
      stop("\n Number of total predecessors (", totalPredecessors,
           ") and total successors (",totalSuccessors,") in the PDG are not equal.")
    }
    if(mergeRate==0.0){ # Next condition only is satisfied when no merging is carried out, since
      # the merge operation may cause having orphan parameters that are removed and in that case
      # the sum of reached cases may not match the reached cases in the root.
      n1 = sumCasesParameterLeaves/numLeafVariables()
      n2 = getRoot()$getParameterAt(1)$numReachedCases()
      if (sumCasesParameterLeaves/numLeafVariables()!=getRoot()$getParameterAt(1)$numReachedCases()){
        stop("Sum of reached cases in leaves for a variable (", n1,")","does not match the reached cases in the root (",n2,")")
      }
    }
    v=c()
    for(p in .self$getParameters()){
      if(p$isContinuous()){
        v = c(v, p$getDensityType() )
      }
    }
    if(length(unique(v))>1){
      print(v)
      stop("There are different types of densities in the PDG.")

    }





    #cat("\n The PDG is valid.")
}
)



##################################################################################################

# PDGClassifier$methods (
#   getIncludedVariablesInPDG = function () {
#     includedVars = list(getRoot())
#     for (var in getVariables()){
#       if(var$hasParent()){
#         includedVars[[length(includedVars)+1]] <- var
#       }
#     }
#     return(includedVars)
#   }
# )
##################################################################################################
# PDGClassifier$methods (
#   getNonIncludedVariablesInPDG = function () {
#     nonIncludedVars = list()
#     for (var in getVariables()[-c(classIndex)]){
#       if(!var$hasParent()){
#         nonIncludedVars[[length(nonIncludedVars)+1]] <- var
#       }
#     }
#     return(nonIncludedVars)
#   }
# )




