#' Function for building a classification(i.e. confusion) matrix for comparing the predictions obtained by
#' a classifier on a set of observations with the real class labels associated to each
#' one of them.
#'
#' @param class_labels a vector containing all class labels.
#'
#' @param real a vector with the real class labels associated to observations.
#'
#' @param predicted a vector with the class labels, predicted by a classifier, for observations.
#'
#' @return Returns a classification matrix summarizing the classification results obtained for each class.
#'
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#'
classificationMatrix <- function (class_labels, real, predicted) {

  # Checks input arguments
  if ((!all(real %in% class_labels)) || (!all(predicted %in% class_labels)))
    stop("Mismatch on class labels")

  if(length(real)!=length(predicted))
    stop("Supplied real classification values (", length(real), ") and predicted values (", length(predicted), ") length difer")

  # Builds the classification matrix
  ncla <- length(class_labels)
  cm <- matrix (nrow = ncla, ncol = ncla)
  for (i in seq_along(class_labels))
    for (j in seq_along(class_labels))
      cm[i,j] <- as.integer(length(which(real[]==class_labels[i] & predicted[]==class_labels[j])))

  rownames(cm) <- class_labels # real class
  colnames(cm) <- class_labels # predicted class

  #class(cm) <- c("classificationMatrix", "matrix")
  #class(cm) <- "classificationMatrix"
  return(cm)
}

#' Function to test whether the matrix contains integer values, has the same number of
#' rows and cols and, row and col names coincide
#'
#' @param m a matrix
#'
#' @return Returns 'TRUE' if matrix contains integer values, has the same number of
#' rows and cols and, row and col names coincide
#'
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' is.classificationMatrix(cm)
#'
is.classificationMatrix <- function (m) {
  if (is.matrix(m))
    if (is.integer(m))
      if (all(rownames(m)==colnames(m)))
         return (TRUE)
  return (FALSE)
}


tp <- function (class_label, classificMatrix) {
  # Checks input arguments
  if (!(class_label %in% rownames(classificMatrix)))
    stop ("Not valid class label")
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  return (classificMatrix[class_label, class_label])
}


fn <- function (class_label, classificMatrix) {
  # Checks input arguments
  if (!(class_label %in% rownames(classificMatrix)))
    stop ("Not valid class label")
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  return (sum(classificMatrix[class_label, which(colnames(classificMatrix) != class_label)]))
}


fp <- function (class_label, classificMatrix) {
  # Checks input arguments
  if (!(class_label %in% rownames(classificMatrix)))
    stop ("Not valid class label")
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  return (sum(classificMatrix[which(rownames(classificMatrix) != class_label), class_label]))
}


tn <- function (class_label, classificMatrix) {
  # Checks input arguments
  if (!(class_label %in% rownames(classificMatrix)))
    stop ("Not valid class label")
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  return (sum(classificMatrix[which(rownames(classificMatrix)!=class_label),which(colnames(classificMatrix)!=class_label)]))
}

#' Gets the proportion of positives (i.e. observations belonging to the class)
#' that are correctly identified as such. It quantifies the avoiding of false negatives.
#' For more than two classes, these results are calculated comparing each class to the remaining
#' classes (i.e. a "one versus all" approach).
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the proportion of true positives for the given class.
#'
#' @seealso \link{classificationMatrix}, \link{fpr}, \link{precision}, \link{fmeasure}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' tprv <- tpr("a", cm)
#'
tpr <- function (class_label, classificMatrix) {
  tpv <- tp(class_label, classificMatrix)
  fnv <- fn(class_label, classificMatrix)
  tprv <- tpv/(tpv+fnv)
  return (tprv)
}

#' Gets the proportion of false positives (i.e. observations that were incorrectly recognized
#' as belonging to the class). For more than two classes, these results are calculated comparing
#' each class to the remaining classes (i.e. a "one versus all" approach).
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the proportion of false positives for the given class.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{precision}, \link{fmeasure}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' fprv <- fpr("a", cm)
#' fprv <- fpr(cl[1], cm)
#'
fpr <- function (class_label, classificMatrix) {
  fpv <- fp(class_label, classificMatrix)
  tnv <- tn(class_label, classificMatrix)
  fprv <- fpv/(fpv+tnv)
  return (fprv)
}

#' Gets the proportion of true positives from observations classified
#' as belonging to the class. For more than two classes, these results
#' are calculated comparing each class to the remaining classes (i.e.
#' a "one versus all" approach).
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the precision for the given class.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{fpr}, \link{fmeasure}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' prev <- precision(cl[1], cm)
#' prev <- precision("a", cm)
#'
precision <- function (class_label, classificMatrix) {
  tpv <- tp(class_label, classificMatrix)
  fpv <- fp(class_label, classificMatrix)
  prev <- tpv/(tpv+fpv)
  return (prev)
}

#' Computes the harmonic mean of sensitivity (i.e. true positive rate) and precision.
#' For more than two classes, these results are calculated comparing each class to
#' the remaining classes (i.e. a "one versus all" approach).
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the f-measure for the given class.
#'
#' @seealso \link{classificationMatrix},\link{tpr}, \link{fpr}, \link{precision}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' fv <- fmeasure("a", cm)
#' fv <- fmeasure(cl[1], cm)
#'
fmeasure <- function (class_label, classificMatrix) {
  tprv <- tpr(class_label, classificMatrix)
  prev <- precision(class_label, classificMatrix)
  fmv  <- 2*tprv*prev/(tprv+prev)
  return (fmv)
}

#' Gets the proportion of negatives (i.e. observations not belonging to the class)
#' that are correctly identified as such. It quantifies the avoiding of false positives.
#' For more than two classes, these results are calculated comparing each class to the
#' remaining classes (i.e. a "one versus all" approach).
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the specificity for the given class.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{fpr}, \link{precision}, \link{fmeasure}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' spev <- specificity("a", cm)
#' spev <- specificity (cl[1], cm)
#'
specificity <- function (class_label, classificMatrix) {
  fpv <- fp(class_label, classificMatrix)
  tnv <- tn(class_label, classificMatrix)
  spev <- tnv/(fpv+tnv)
  return (spev)
}

#' Computes the area under the roc curve (AUC) using the trapezoid approximation
#' on a single (TPR, FPR) point.
#'
#' @param class_label the label of the class.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the AUC for the given class.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{fpr}, \link{precision}, \link{fmeasure}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' aucv <- auc("a", cm)
#' aucv <- auc(cl[1], cm)
#'
auc <- function (class_label, classificMatrix) {
  tprv <- tpr(class_label, classificMatrix)
  spev <- specificity(class_label, classificMatrix)
  aucv <- (tprv+spev)/2
  return (aucv)
}

#' Gets the proportion of observations correctly classified.
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns the proportion of correctly classified observations.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{fpr}, \link{precision}, \link{fmeasure}, \link{specificity}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' tprv <- tpr("a", cm)
#' fprv <- fpr(cl[1], cm)
#' prev <- precision("a", cm)
#' fv <- fmeasure(cl[1], cm)
#' spev <- specificity (cl[1], cm)
#' accv <- accuracy(cm)
#'
accuracy <- function (classificMatrix) {
  # Checks input argument
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  # Computes accuracy
  cci <- sum(diag(classificMatrix))
  tci <- sum(classificMatrix)
  acc <- cci / tci
  return (acc)
}

#' Sumarizes the classification results encoded in a classification matrix. Detiled results by class
#' are included. For more than two classes, these results are calculated comparing each class to the
#' remaining classes (i.e. a "one versus all" approach).
#'
#' @param classificMatrix a classification matrix.
#'
#' @return Returns a summary of the classifications results, including the confusion matrix itself,
#' detailed results by class and global accuracy.
#'
#' @seealso \link{classificationMatrix}, \link{tpr}, \link{fpr}, \link{precision}, \link{fmeasure}, \link{specificity}, \link{accuracy}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' summarize(cm)
#'
summarize <- function (classificMatrix) {
  # Checks input argument
  # if (class(classificMatrix)!="classificationMatrix")
  #  stop ("Object must be of class 'classificationMatrix'")

  # Computes the summary
  cat("\nClassification Matrix (rows: real classes, cols: predicted classes)\n\n")
  show (classificMatrix)
  cat ("\n")

  cat("\nDetailed Accuracy By Class\n\n")
  class_labels <- rownames(classificMatrix)
  TPR <- sapply (class_labels, tpr, classificMatrix)
  FPR <- sapply (class_labels, fpr, classificMatrix)
  Precision <- sapply (class_labels, precision, classificMatrix)
  f_measure <- sapply (class_labels, fmeasure, classificMatrix)
  results <- cbind(TPR, FPR, Precision, f_measure)
  show(results)
  cat ("\n")
  cat ("Weigthed Average \n")
  weight <- rowSums(classificMatrix)
  weight <- weight / sum(weight)
  weigthedAvgResults <- apply (results, 2, stats::weighted.mean, weight)
  show(weigthedAvgResults)
  cat("\n")
  cat("Percentage of Correctly Classified Instances",accuracy(classificMatrix)*100,"%\n\n")
}

#' Plots the classes counts in a classification matrix.
#'
#' @param classificMatrix a confusion matrix.
#'
#' @param short indicates the type of plot. By default a short plot is produced: observations are grouped by class.
#'
#' @return Returns a plot of the class counts in a confusion matrix.
#'
#' @seealso \link{classificationMatrix}
#' @export
#' @examples
#' ## Builds a classification matrix
#' cl <- c("a", "b", "c")
#' rl <- c("a", "b", "b", "c", "a", "a")
#' pl <- c("a", "b", "c", "c", "b", "a")
#' cm <- classificationMatrix(cl, rl, pl)
#' cmplot(cm)
#' cmplot(cm, FALSE)
#' cmplot(cm, short=FALSE)
#'
cmplot <- function (classificMatrix, short=TRUE) {
  # Checks input argument
  if (!is.classificationMatrix(classificMatrix))
    stop ("Object must be a 'classificationMatrix'")

  # Plots the confusion matrix
  mo <- max(classificMatrix) # maximum number of observations
  if (short) {
    # as a single graph; observations are grouped by class
    bp <- graphics::barplot (classificMatrix, beside = T, xlab="classes", ylab="observations", ylim=c(0,mo))
    graphics::text (bp, 0, classificMatrix, cex=1, pos=3)
  } else {
    # combining a graph by class
    nc <- length(rownames(classificMatrix))
    graphics::par(mfrow = c(nc, 1))
    for (i in 1:nc) { #seq_along(rownames(cm))) {
      bp <- graphics::barplot(classificMatrix[i,], xlab=paste("class ",rownames(classificMatrix)[i]," as"), ylab="observations", ylim=c(0,mo))
      graphics::text (bp, 0, classificMatrix[i,], cex=1, pos=3)
    }
  }
}

#' Plots a ROC curve from a series of labels and scores returned by a classifier.
#' In multiclass cases plots a ROC curve for each class applying a one versus all
#' approach.
#'
#' @param real a vector with the real class labels associated to observations.
#'
#' @param score a vector with the probability values assigned to each observation
#' by the classifier. The length of this vector has to coincid with that of the real
#' labels vector.
#'
#' @param single an optional boolean. It indicates if a single plot has to be
#' produced or not. By default a single plot is produced.
#'
#' @param plot_classes an optional vector with the different class labels whose
#' ROC curves have to be displayed. Only those classes included in the \code{real}
#' vector are taken into account. By default it is set to NULL.
#'
#' @return Returns plots with one ROC curve (binary case) or with a ROC curve per
#' class (multiclass case) using the the class reference formulation (i.e. one
#' versus all) approach. Classes are considered in the order specified by
#' the vector with the \code{real} class labels.
#'
#' @seealso \link{classificationMatrix}, \code{pROC}
#' @export
#' @examples
#'
#' ## Simulates classification results
#' class_values <- sample(c(TRUE, FALSE), 20, replace = TRUE )
#' classifier_scores <- runif(20)
#' ROCplot(class_values, classifier_scores)
#'
#' ## Simulates multiclASS classification results
#' class_values <- sample( c("a","b","c"), 20, replace = TRUE )
#' classifier_scores <- runif(20)
#' # Plot multiclass ROC
#' # A unique plot
#' ROCplot(class_values, classifier_scores)
#' ROCplot(class_values, classifier_scores, plot_classes = c("a", "c"))
#' # One plot per class
#' ROCplot(class_values, classifier_scores, FALSE)
#' ROCplot(class_values, classifier_scores, FALSE, plot_classes = c("a", "c"))
#'
ROCplot <- function (real, score, single = TRUE, plot_classes = NULL) {
  # Checks input arguments
  if(length(real)!=length(score))
    stop("Supplied real classification values (", length(real), ") and classifier score values (", length(score), ") length difer")

  ctag <- unique(real)# class labels in real
  ncr <- length(ctag) # number of classes in real

  if (ncr > 2) {
    # Define a color palette for multiple plot
    colorpalette <- grDevices::rainbow(ncr)

    # multiclass ROC plot
    roc_data <- pROC::multiclass.roc(real, score)


    if (single == TRUE) {

      # Produce one multiclass ROC plot

      if (is.null(plot_classes)) {
        # all clases in real are taken into account
        # produce a multiROC plot
        # create the base ROC plot
        graphics::plot(roc_data[['rocs']][[1]], legacy.axes=TRUE, col=colorpalette[1], xlab ="FPR", ylab ="TPR")

        for (i in 2:ncr) {
          # add subsequent ROC plots
          graphics::plot(roc_data[['rocs']][[i]], add=TRUE, legacy.axes=TRUE, col=colorpalette[i], xlab ="FPR", ylab ="TPR")
        }
        graphics::legend('bottomright', inset=.02, ctag, lty=1, col=colorpalette,  cex=.75)
      } else {
        # only the classes in plot_classes are taken into account
        if (all(plot_classes %in% ctag)) {
          pci <- which(ctag %in% plot_classes)  #indexes of the classes to plot

          # produce a multiROC plot
          # create the base ROC plot
          graphics::plot(roc_data[['rocs']][[pci[1]]], legacy.axes=TRUE, col=colorpalette[pci[1]], xlab ="FPR", ylab ="TPR")

          for (i in pci[-1]) {
            # add subsequent ROC plots
            graphics::plot(roc_data[['rocs']][[i]], add=TRUE, legacy.axes=TRUE, col=colorpalette[i], xlab ="FPR", ylab ="TPR")
          }
          graphics::legend('bottomright', inset=.02, ctag[pci], lty=1, col=colorpalette[pci],  cex=.75)
        }
      }

    } else {

      # Produce a series of individual ROC plots for all classes in real

      if (is.null(plot_classes)) {
        for (i in 1:ncr) {
          # produce one individual ROC plot per class in real
          graphics::plot(roc_data[['rocs']][[i]], legacy.axes=TRUE, col=colorpalette[i], xlab ="FPR", ylab ="TPR")
          graphics::legend('bottomright', inset=.02, ctag[i], lty=1, col=colorpalette[i],  cex=.75)
        }
      } else {
        # produce individual ROC plots for classes in plot_classes
        if (all(plot_classes %in% ctag)) {
          pci <- which(ctag %in% plot_classes)  #indexes of the classes to plot

          # produce one ROC plot for each class
          for (i in pci) {
            # add subsequent ROC plots
            graphics::plot(roc_data[['rocs']][[i]], legacy.axes=TRUE, col=colorpalette[i], xlab ="FPR", ylab ="TPR")
            graphics::legend('bottomright', inset=.02, ctag[i], lty=1, col=colorpalette[i],  cex=.75)
          }
        }
      }

    }

  } else {

    # ROC plot
    roc_data <- pROC::roc(real, score)
    graphics::plot(roc_data, legacy.axes=TRUE, col="red", xlab ="FPR", ylab ="TPR")
    #graphics::legend('bottomright', inset=.02, ctag[1], lty=1, col=colorpalette[1],  cex=.75)
  }
}



